import os
from random import getrandbits

from flask import request, current_app, Response
from flask_redis import FlaskRedis
from nidhoggr.core.repository import BaseUserRepo
from nidhoggr.core.user import User
from werkzeug.exceptions import BadRequest

from valhalla.decorators import as_json, MIME_TYPE
from valhalla.jwt import generate_auth_token
from valhalla.models import (
    AuthFirstPhaseRequest,
    AuthFirstPhaseResponse,
    AuthSecondPhaseRequest,
    AuthSecondPhaseResponse,
    AuthCookie
)
from valhalla.utils import validate_form, handle_status

SERVER_ID = os.getenv('SERVER_ID')
AUTH_COOKIE_EXPIRE = int(os.getenv('AUTH_COOKIE_EXPIRE'))


@as_json
def handshake() -> AuthFirstPhaseResponse:
    redis: FlaskRedis = current_app.redis

    req = handle_status(validate_form(form=request.form, clazz=AuthFirstPhaseRequest))
    token = getrandbits(32)

    redis.set(req.name, AuthCookie(ip=request.remote_addr, token=token).json(), ex=AUTH_COOKIE_EXPIRE)

    return AuthFirstPhaseResponse(
        offline=False,  # Don't support offline mode
        serverId=SERVER_ID,
        verifyToken=token
    )


def response():
    redis: FlaskRedis = current_app.redis
    users: BaseUserRepo = current_app.users

    req = handle_status(validate_form(form=request.form, clazz=AuthSecondPhaseRequest))

    cookie_raw = redis.get(req.name)
    redis.delete(req.name)

    if not cookie_raw:
        raise BadRequest(description='Invalid or expired cookie, try again')

    cookie = AuthCookie.parse_raw(cookie_raw)

    if cookie.ip != request.remote_addr:
        raise BadRequest(description='Unauthenticated network')
    if cookie.token != req.verifyToken:
        raise BadRequest(description='Wrong token')

    user: User = handle_status(users.get_user(login=req.name))

    if req.serverId != user.server:
        raise BadRequest(description='Unauthenticated login')

    resp_payload = AuthSecondPhaseResponse(accessToken=user.access, userId=user.uuid)
    resp = Response(
        mimetype=MIME_TYPE,
        content_type=MIME_TYPE,
        response=resp_payload.json(exclude_none=True)
    )
    resp.headers['Authorization'] = f"Bearer {generate_auth_token(uuid=user.uuid).decode('ascii')}"
    return resp



from base64 import b64encode
from datetime import datetime
from typing import Union
from uuid import UUID

from flask import current_app, request
from nidhoggr.core.repository import BaseTextureRepo, BaseUserRepo
from nidhoggr.core.response import TextureStatusResponse
from nidhoggr.core.texture import ComplexTextureResponse, TextureRequest, TextureType, TextureUploadRequest
from nidhoggr.core.user import User

from valhalla.decorators import as_json, protected
from valhalla.utils import handle_status, validate_image, validate_metadata


@as_json
def fetch(uuid: UUID) -> Union[TextureStatusResponse, ComplexTextureResponse]:
    textures: BaseTextureRepo = current_app.textures
    users: BaseUserRepo = current_app.users

    user: User = handle_status(users.get_user(uuid=uuid))

    if user.synthetic:
        return TextureStatusResponse(message=f'There are no textures for synthetic user with uuid {uuid}')

    texture_request = TextureRequest(uuid=uuid, texture_types=list(TextureType))
    texture_response = handle_status(textures.get(request=texture_request))

    return ComplexTextureResponse(
        timestamp=int(datetime.now().timestamp() * 1000),
        profileId=user.uuid,
        profileName=user.login,
        textures=texture_response.textures
    )


@protected
@as_json
def put(uuid: UUID, kind: TextureType) -> TextureStatusResponse:
    image = request.files.get('file')
    metadata = dict(request.form)

    validate_image(raw_image=image)
    validate_metadata(raw_metadata=metadata)

    textures: BaseTextureRepo = current_app.textures

    upload_request = TextureUploadRequest(
        uuid=uuid,
        data=b64encode(image),
        kind=kind,
        metadata=metadata
    )

    return handle_status(textures.upload(request=upload_request))


@protected
@as_json
def post(uuid: UUID, kind: TextureType) -> TextureStatusResponse:
    return TextureStatusResponse(message="Web uploads disabled due to security reasons")


@protected
@as_json
def delete(uuid: UUID, kind: TextureType) -> TextureStatusResponse:
    textures: BaseTextureRepo = current_app.textures
    delete_request = TextureRequest(uuid=uuid, texture_types=[kind])

    return handle_status(textures.clear(request=delete_request))



import os
from io import BytesIO
from typing import Union, TypeVar, Any, Dict, Type

from PIL import Image
from nidhoggr.core.response import ErrorResponse, TextureStatusResponse
from pydantic import BaseModel, ValidationError
from werkzeug.datastructures import ImmutableMultiDict
from werkzeug.exceptions import InternalServerError, BadRequest

from valhalla.decorators import as_json

T = TypeVar("T")
U = TypeVar("U", bound=BaseModel)

TOKEN_EXPIRE = os.getenv('TOKEN_EXPIRE')
SECRET_KEY = os.getenv('SECRET_KEY')


def handle_status(repository_response: Union[ErrorResponse, T]) -> T:
    if isinstance(repository_response, ErrorResponse):
        raise InternalServerError(description=repository_response.reason)
    return repository_response


@as_json
def error_handler(error: InternalServerError):
    return TextureStatusResponse(message=error.description)


def validate_image(*, raw_image: bytes):
    if not raw_image:
        raise BadRequest(description=f'Empty image')

    with Image.open(BytesIO(raw_image)) as image:
        if image.format != "PNG":
            raise BadRequest(description=f"Format not allowed: {image.format}")

        sizes = {64, 128, 256, 512, 1024, 2048}
        (width, height) = image.size
        valid = width / 2 == height or width == height

        if not valid or width not in sizes:
            raise BadRequest(description=f"Unsupported image size: {image.size}")


def validate_metadata(*, raw_metadata: Dict[Any, Any]):
    if not all((
        isinstance(k, str) and isinstance(v, str)
        for k, v
        in raw_metadata
    )):
        raise BadRequest(description=f"Malformed metadata")


def validate_form(*, form: ImmutableMultiDict, clazz: Type[U]) -> Union[ErrorResponse, U]:
    try:
        return clazz.parse_obj(form.to_dict())
    except ValidationError as e:
        raise BadRequest(description=str(e))




from typing import Optional
from uuid import UUID
from flask_httpauth import HTTPTokenAuth

from itsdangerous import TimedJSONWebSignatureSerializer, SignatureExpired, BadSignature

from valhalla.utils import SECRET_KEY, TOKEN_EXPIRE
from werkzeug.exceptions import Unauthorized


auth = HTTPTokenAuth()


def generate_auth_token(*, uuid: UUID):
    serializer = TimedJSONWebSignatureSerializer(SECRET_KEY, expires_in=TOKEN_EXPIRE)
    return serializer.dumps({'uuid': uuid.hex})


@auth.error_handler
def auth_failed():
    raise Unauthorized(description="Authentication failed")


@auth.verify_token
def verify_auth_token(token) -> Optional[UUID]:
    serializer = TimedJSONWebSignatureSerializer(SECRET_KEY)
    try:
        data = serializer.loads(token)
    except SignatureExpired:
        return None
    except BadSignature:
        return None
    else:
        return UUID(data['id'])



from flask import Flask
from nidhoggr.core.texture import TextureType
from werkzeug.routing import BaseConverter, ValidationError


class TextureTypeConverter(BaseConverter):
    def to_python(self, value):
        try:
            return TextureType(value.upper())
        except ValueError as e:
            raise ValidationError(f"Failed to parse {value} as registered texture type") from e

    def to_url(self, value):
        return str(value)


class FlaskTextureType:
    """Flask extension providing a Texture type url converter"""
    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        app.url_map.converters['kind'] = TextureTypeConverter



from functools import wraps
from typing import Callable

from flask import Response
from pydantic import BaseModel
from werkzeug.exceptions import NotFound, Forbidden

from valhalla.jwt import auth

MIME_TYPE = 'application/json'


def as_json(func: Callable):
    @wraps(func)
    def wrapper(*view_args, **view_kwargs) -> Response:
        response: BaseModel = func(*view_args, **view_kwargs)
        return Response(
            mimetype=MIME_TYPE,
            content_type=MIME_TYPE,
            response=response.json(exclude_none=True)
        )
    return wrapper


def protected(func: Callable):
    @wraps(func)
    def wrapper(*view_args, **view_kwargs) -> Response:
        wrapped = auth.login_required(func)
        current_uuid = auth.current_user()
        if current_uuid is None:
            raise NotFound(description=f"There are no such user")
        if current_uuid != view_args[0]:  # By convention
            raise Forbidden(description=f"Cannot change another user's textures")
        return wrapped(*view_args, **view_kwargs)
    return wrapper



import os

from flask import Flask
from flask_redis import FlaskRedis
from flask_uuid import FlaskUUID
from nidhoggr.core.repository import BaseTextureRepo, BaseUserRepo
from werkzeug.exceptions import InternalServerError, BadRequest, Unauthorized, NotFound, Forbidden

from valhalla import utils
from valhalla.views import auth, texture
from valhalla.converter import FlaskTextureType


def configure_error_handlers(app: Flask):
    app.register_error_handler(InternalServerError, utils.error_handler)
    app.register_error_handler(BadRequest, utils.error_handler)
    app.register_error_handler(Unauthorized, utils.error_handler)
    app.register_error_handler(NotFound, utils.error_handler)
    app.register_error_handler(Forbidden, utils.error_handler)


def configure_views(app: Flask):
    app.add_url_rule('/api/v1/auth/handshake', 'auth_handshake', auth.handshake, methods=['POST'])
    app.add_url_rule('/api/v1/auth/response', 'auth_response', auth.response, methods=['POST'])
    app.add_url_rule('/api/v1/user/<uuid:uuid>', 'texture_fetch', texture.fetch, methods=['GET'])
    app.add_url_rule('/api/v1/user/<uuid:uuid>/<kind:kind>', 'texture_post', texture.post, methods=['POST'])
    app.add_url_rule('/api/v1/user/<uuid:uuid>/<kind:kind>', 'texture_put', texture.put, methods=['PUT'])
    app.add_url_rule('/api/v1/user/<uuid:uuid>/<kind:kind>', 'texture_delete', texture.delete, methods=['DELETE'])


def configure_redis(app: Flask, redis: FlaskRedis):
    app.config["REDIS_URL"] = os.getenv("REDIS_URL")
    redis.init_app(app)
    app.redis = redis


def create_app(textures: BaseTextureRepo, users: BaseUserRepo, redis: FlaskRedis) -> Flask:
    app = Flask(__package__, root_path=os.path.dirname(__file__))
    FlaskUUID(app)
    FlaskTextureType(app)
    app.users = users
    app.textures = textures
    configure_error_handlers(app)
    configure_views(app)
    configure_redis(app, redis)
    return app



from pydantic import BaseModel, UUID4


class AuthFirstPhaseRequest(BaseModel):
    name: str

    class Config:
        allow_mutation = False


class AuthFirstPhaseResponse(BaseModel):
    offline: bool
    serverId: str
    verifyToken: int

    class Config:
        allow_mutation = False


class AuthSecondPhaseRequest(AuthFirstPhaseRequest):
    verifyToken: int

    class Config:
        allow_mutation = False


class AuthSecondPhaseResponse(BaseModel):
    accessToken: UUID4
    userId: UUID4

    class Config:
        allow_mutation = False


class AuthCookie(BaseModel):
    ip: str
    token: int

    class Config:
        allow_mutation = False
